/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// r_light.c

#include "quakedef.h"


/*
==================
R_AnimateLight
==================
*/
void R_AnimateLight (double time)
{
	// 'm' is normal light, 'a' is no light, 'z' is double bright
	int flight = (int) floor (time * 10.0);
	int clight = (int) ceil (time * 10.0);
	float lerpfrac = (float) ((time * 10.0) - flight);
	int i;

	float lightmul = ((float) 'z' - (float) 'a') / ((float) 'm' - (float) 'a');

	for (i = 0; i < MAX_LIGHTSTYLES; i++)
	{
		float light = 'm';

		if (cl_lightstyle[i].length && r_dynamic.value)
		{
			if (cl_lightstyle[i].length == 1)
				light = cl_lightstyle[i].map[0];
			else
			{
				light = (
					(float) (cl_lightstyle[i].map[flight % cl_lightstyle[i].length]) * (1.0f - lerpfrac) +
					(float) (cl_lightstyle[i].map[clight % cl_lightstyle[i].length]) * lerpfrac
				);
			}
		}

		// bring back to normal light range and scale for overbrighting
		//d_lightstylevalue[i] = (light - (float) 'a') / ((float) 'z' - (float) 'a') * lightmul;
		d_lightstylevalue[i] = (light - (float) 'a') / ((float) 'm' - (float) 'a');
	}
}


/*
=============================================================================

DYNAMIC LIGHTS

=============================================================================
*/

/*
=============
R_MarkLights
=============
*/
void R_MarkLights (dlight_t *light, int bit, mnode_t *node)
{
}


/*
=============
R_PushDlights
=============
*/
void R_PushDlights (mnode_t *headnode, glmatrix *localmatrix)
{
}


/*
=============================================================================

LIGHT SAMPLING

=============================================================================
*/

mplane_t		*lightplane;
vec3_t			lightspot;

int RecursiveLightPoint (mnode_t *node, vec3_t start, vec3_t end)
{
	return 0;
}

int R_LightPoint (vec3_t p)
{
	return 0;
}


staticlight_t r_staticlights[MAX_MAP_ENTITIES];
int r_numstaticlights = 0;

int		numlighttargets;
char	lighttargets[32][64];

int LightStyleForTargetname (char *targetname, qboolean alloc)
{
	int	i;
	
	for (i = 0; i < numlighttargets; i++)
		if (!strcmp (lighttargets[i], targetname))
			return 32 + i;

	if (!alloc)
		return -1;

	strcpy (lighttargets[i], targetname);
	numlighttargets++;

	return numlighttargets - 1 + 32;
}


void MatchTargets (void)
{
	int	i, j;

	for (i = 0; i < r_numstaticlights; i++)
	{
		if (!r_staticlights[i].target[0])
			continue;
			
		for (j = 0; j < r_numstaticlights; j++)
		{
			if (!strcmp (r_staticlights[j].targetname, r_staticlights[i].target))
			{
				r_staticlights[i].targetent = &r_staticlights[j];
				break;
			}
		}

		if (j == r_numstaticlights)
		{
			Con_DPrintf (
				"WARNING: entity at (%i,%i,%i) (%s) has unmatched target\n",
				(int) r_staticlights[i].origin[0],
				(int) r_staticlights[i].origin[1],
				(int) r_staticlights[i].origin[2],
				r_staticlights[i].classname
			);

			continue;
		}

		// set the style on the source ent for switchable lights
		if (r_staticlights[j].style)
		{
			r_staticlights[i].style = r_staticlights[j].style;
		}
	}
}


void R_LoadLights (void)
{
	char *data = cl.worldmodel->entities;
	staticlight_t *staticlight;
	char		key[64];
	double		vec[3];
	int	i;

	memset (r_staticlights, 0, sizeof (r_staticlights));
	r_numstaticlights = 0;

	// go through all the entities
	while (1)
	{
		// parse the opening brace	
		data = COM_Parse (data);

		if (!data)
			break;

		if (com_token[0] != '{')
			Sys_Error ("R_LoadLights: found %s when expecting {", com_token);

		if (r_numstaticlights == MAX_MAP_ENTITIES)
			Sys_Error ("R_LoadLights: MAX_MAP_ENTITIES");

		staticlight = &r_staticlights[r_numstaticlights];

		// go to a new static light
		// note - we need to also load non-light entities because we need targetents for spotlights
		r_numstaticlights++;

		// go through all the keys in this entity
		while (1)
		{
			int	c;

			// parse key
			data = COM_Parse (data);

			if (!data)
				Sys_Error ("R_LoadLights: EOF without closing brace");

			if (!strcmp (com_token, "}"))
				break;

			strcpy (key, com_token);

			// parse value
			data = COM_Parse (data);

			if (!data)
				Sys_Error ("R_LoadLights: EOF without closing brace");

			c = com_token[0];

			if (c == '}')
				Sys_Error ("R_LoadLights: closing brace without data");

			if (!strcmp (key, "classname"))
				strcpy (staticlight->classname, com_token);
			else if (!strcmp (key, "target"))
				strcpy (staticlight->target, com_token);
			else if (!strcmp (key, "targetname"))
				strcpy (staticlight->targetname, com_token);
			else if (!strcmp (key, "origin"))
			{
				// scan into doubles, then assign
				// which makes it vec_t size independent
				if (sscanf (com_token, "%lf %lf %lf", &vec[0], &vec[1], &vec[2]) != 3)
					Sys_Error ("R_LoadLights: not 3 values for origin");

				for (i = 0; i < 3; i++)
					staticlight->origin[i] = vec[i];
			}
			else if (!strncmp (key, "light", 5) || !strcmp (key, "_light"))
			{
				staticlight->light = atof (com_token);
			}
			else if (!strcmp (key, "style"))
			{
				staticlight->style = atof (com_token);
				if ((unsigned) staticlight->style > 254)
					Sys_Error ("Bad light style %i (must be 0-254)", staticlight->style);
			}
			else if (!strcmp (key, "angle"))
			{
				staticlight->angle = atof (com_token);
			}
		}

		// all fields have been parsed
		if (!strncmp (staticlight->classname, "light", 5) && !staticlight->light)
			staticlight->light = DEFAULTLIGHTLEVEL;

		if (!strcmp (staticlight->classname, "light"))
		{
			if (staticlight->targetname[0] && !staticlight->style)
			{
				staticlight->style = LightStyleForTargetname (staticlight->targetname, true);
			}
		}

		if (staticlight->light > 0)
		{
			CL_DLightColorScale (staticlight->color, DL_COLOR_WHITE);
			R_AddLfrags (staticlight);
			// add leaf fragments for this light
		}
	}

	Con_DPrintf ("%d lights read\n", r_numstaticlights);

	MatchTargets ();
}


void R_MarkLights_PointLight (dlight_t *dl, mnode_t *node)
{
	mplane_t	*splitplane;
	float		dist;
	msurface_t	*surf;
	int			i, sidebit;
	float maxdist = dl->radius * dl->radius;

	if (node->contents < 0)
		return;

	splitplane = node->plane;
	dist = DotProduct (dl->transformed, splitplane->normal) - splitplane->dist;

	if (dist > dl->radius)
	{
		R_MarkLights_PointLight (dl, node->children[0]);
		return;
	}

	if (dist < -dl->radius)
	{
		R_MarkLights_PointLight (dl, node->children[1]);
		return;
	}

	// mark the polygons
	surf = node->surfaces;

	for (i = 0; i < node->numsurfaces; i++, surf++)
	{
		int s, t;
		float impact[3], l;

		if (surf->visframe != r_framecount) continue;
		if (surf->flags & (SURF_DRAWTURB | SURF_DRAWSKY)) continue;

		// ensures surfaces are only added once
		if (surf->dlightframe == r_dlightframecount) continue;

		// skip surfaces facing away from the light (fixme - is this not the same plane as above?)
		dist = DotProduct (dl->transformed, surf->plane->normal) - surf->plane->dist;

		if (dist >= 0)
			sidebit = 0;
		else sidebit = SURF_PLANEBACK;

		if ((surf->flags & SURF_PLANEBACK) != sidebit)
			continue;

		impact[0] = dl->transformed[0] - surf->plane->normal[0] * dist;
		impact[1] = dl->transformed[1] - surf->plane->normal[1] * dist;
		impact[2] = dl->transformed[2] - surf->plane->normal[2] * dist;

		// clamp center of light to corner and check brightness
		l = Vector3Dot (impact, surf->texinfo->vecs[0]) + surf->texinfo->vecs[0][3] - surf->texturemins[0];
		s = l + 0.5; if (s < 0) s = 0; else if (s > surf->extents[0]) s = surf->extents[0];
		s = l - s;

		l = Vector3Dot (impact, surf->texinfo->vecs[1]) + surf->texinfo->vecs[1][3] - surf->texturemins[1];
		t = l + 0.5; if (t < 0) t = 0; else if (t > surf->extents[1]) t = surf->extents[1];
		t = l - t;

		// compare to minimum light
		if ((s * s + t * t + dist * dist) < maxdist)
		{
			// chain the surface
			surf->texturechain = surf->texinfo->texture->texturechain;
			surf->texinfo->texture->texturechain = surf;

			surf->dlightframe = r_dlightframecount;
			dl->numsurfaces++;
		}
	}

	R_MarkLights_PointLight (dl, node->children[0]);
	R_MarkLights_PointLight (dl, node->children[1]);
}


void R_LightEntity (dlight_t *dl, entity_t *ent, spotparms_t *spotparms)
{
	model_t *clmodel = ent->model;
	float dist[3];

	// go to a new dlight frame
	r_dlightframecount++;

	// move the light to the entity frame of reference
	R_InverseTransform (&ent->localmatrix, dl->transformed, dl->origin);

	// vertex program params
	glProgramEnvParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, dl->transformed);

	switch (clmodel->type)
	{
	case mod_alias:
		// make this a proper sphere/box intersection test
		Vector3Subtract (dist, ent->lerpedorigin, dl->origin);

		if (Vector3Length (dist) < dl->radius)
		{
			if (spotparms->isSpotLight)
				R_DrawAliasModel (ent, r_lightmeshvp, r_spotlightfp);
			else R_DrawAliasModel (ent, r_lightmeshvp, r_pointlightfp);
		}

		break;

	case mod_brush:
		dl->numsurfaces = 0;

		R_ClearTextureChains (clmodel);

		if (clmodel == cl.worldmodel)
			R_MarkLights_PointLight (dl, clmodel->nodes);
		else if (clmodel->firstmodelsurface != 0)
			R_MarkLights_PointLight (dl, clmodel->nodes + clmodel->hulls[0].firstclipnode);
		else R_MarkLights_PointLight (dl, clmodel->nodes);

		if (spotparms->isSpotLight)
			GL_BindPrograms (r_lightsurfvp, r_spotlightfp);
		else GL_BindPrograms (r_lightsurfvp, r_pointlightfp);

		if (dl->numsurfaces)
			R_DrawTextureChains_LightPass (clmodel, &ent->localmatrix);

		break;
	}
}


qboolean R_ScissorLight (float *lightpos, float r)
{
	// https://www.opengl.org/discussion_boards/showthread.php/153499-Is-this-a-correct-way-to-compute-the-scissor-rectangle?p=1098685&viewfull=1#post1098685
	// ugh.  lengyel.  this should be doable by just transforming corners of a bbox from world to screen space.  you would think.
	int sx = vid.width;
	int sy = vid.height - sb_lines; // adjust for sbar
	int rect[4] = {0, 0, sx, sy};
	float d;
	float nx1, nx2, ny1, ny2, nz1, nz2, pz1, pz2;
	float r2, l[3], l2[3], e1, e2;

	r2 = r * r;		// squared radius

	// move the light to eye space
	R_Transform (&r_world_matrix, l, lightpos);

	Vector3Set (l2, l[0] * l[0], l[1] * l[1], l[2] * l[2]);

	// take these from the projection matrix instead of recalculating them a different way
	e1 = r_projection_matrix.m4x4[0][0];
	e2 = r_projection_matrix.m4x4[1][1];

	d = r2 * l2[0] - (l2[0] + l2[2]) * (r2 - l2[2]);

	if (d >= 0)
	{
		d = sqrtf (d);

		nx1 = (r * l[0] + d) / (l2[0] + l2[2]);
		nx2 = (r * l[0] - d) / (l2[0] + l2[2]);

		nz1 = (r - nx1 * l[0]) / l[2];
		nz2 = (r - nx2 * l[0]) / l[2];

		pz1 = (l2[0] + l2[2] - r2) / (l[2] - (nz1 / nx1) * l[0]);
		pz2 = (l2[0] + l2[2] - r2) / (l[2] - (nz2 / nx2) * l[0]);

		if (pz1 >= 0 && pz2 >= 0)
			return false; // discard light

		if (pz1 < 0)
		{
			float fx = nz1 * e1 / nx1;
			int ix = (int) ((fx + 1.0f) * sx * 0.5f);
			float px = -pz1 * nz1 / nx1;

			if (px < l[0])
				rect[0] = max (rect[0], ix);
			else rect[2] = min (rect[2], ix);
		}

		if (pz2 < 0)
		{
			float fx = nz2 * e1 / nx2;
			int ix = (int) ((fx + 1.0f) * sx * 0.5f);
			float px = -pz2 * nz2 / nx2;

			if (px < l[0])
				rect[0] = max (rect[0], ix);
			else rect[2] = min (rect[2], ix);
		}

		if (rect[2] <= rect[0])
			return false; // discard light
	}

	d = r2 * l2[1] - (l2[1] + l2[2]) * (r2 - l2[2]);

	if (d >= 0)
	{
		d = sqrtf (d);

		ny1 = (r * l[1] + d) / (l2[1] + l2[2]);
		ny2 = (r * l[1] - d) / (l2[1] + l2[2]);

		nz1 = (r - ny1 * l[1]) / l[2];
		nz2 = (r - ny2 * l[1]) / l[2];

		pz1 = (l2[1] + l2[2] - r2) / (l[2] - (nz1 / ny1) * l[1]);
		pz2 = (l2[1] + l2[2] - r2) / (l[2] - (nz2 / ny2) * l[1]);

		if (pz1 >= 0 && pz2 >= 0)
			return false; // discard light

		if (pz1 < 0)
		{
			float fy = nz1 * e2 / ny1;
			int iy = (int) ((fy + 1.0f) * sy * 0.5f);
			float py = -pz1 * nz1 / ny1;

			if (py < l[1])
				rect[1] = max (rect[1], iy);
			else rect[3] = min (rect[3], iy);
		}

		if (pz2 < 0)
		{
			float fy = nz2 * e2 / ny2;
			int iy = (int) ((fy + 1.0f) * sy * 0.5f);

			float py = -pz2 * nz2 / ny2;

			if (py < l[1])
				rect[1] = max (rect[1], iy);
			else rect[3] = min (rect[3], iy);
		}

		if (rect[3] <= rect[1])
			return false; // discard light
	}

	// render light with scissor (adjust for sbar)
	glScissor (rect[0], rect[1] + sb_lines, rect[2] - rect[0], rect[3] - rect[1]);

	// the light is on-screen now
	return true;
}


qboolean R_RenderLightAndShadow (dlight_t *dl, float style, spotparms_t *spotparms)
{
	int i;

	// lights can decay or animate to black
	if (!(dl->radius > 0)) return false;
	if (!(style > 0)) return false;

	// lights outside the view frustum
	if (R_CullSphere (dl->origin, dl->radius)) return false;

	// scissor test optimization
	if (!R_ScissorLight (dl->origin, dl->radius)) return false;

	// fragment program params
	glProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 0, dl->radius, dl->radius, dl->radius, 0);

	// coloured light
	//glProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 1, dl->color[0] * style, dl->color[1] * style, dl->color[2] * style, 0);

	// white light
	glProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 1, style, style, style, 0);

	if (spotparms->isSpotLight)
	{
		glProgramEnvParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 2, spotparms->spotvec);
		glProgramEnvParameter4fARB (GL_FRAGMENT_PROGRAM_ARB, 3, spotparms->falloff, 0, 0, 0);

		GL_BindTexture (1, r_spottexture);
	}

	// **********************************
	// shadow
	// **********************************

	// **********************************
	// light
	// **********************************

	// go to additive blending
	GL_BlendState (BS_ADDITIVE);
	GL_DepthState (DS_TESTEQUAL);

	// light the world model
	R_LightEntity (dl, &r_worldentity, spotparms);

	// light entities
	for (i = 0; i < cl_numvisedicts; i++)
	{
		currententity = cl_visedicts[i];
		R_LightEntity (dl, cl_visedicts[i], spotparms);
	}

	// clear stencil buffer
	// mark shadows (worldmodel + entities)
	// mark lights (worldmodel + entities)

	return true;
}


void R_RenderLightsAndShadows (void)
{
	int i;
	spotparms_t spotparms;
	int numlights = 0;

	glEnable (GL_SCISSOR_TEST);

	for (i = 0; i < r_numstaticlights; i++)
	{
		dlight_t dl;
		staticlight_t *light = &r_staticlights[i];

		if (light->visframe != r_framecount) continue;

		// build a dynamic light from this static light
		Vector3Copy (dl.origin, light->origin);
		Vector3Copy (dl.color, light->color);
		dl.minlight = 0;
		dl.radius = light->light;

		if (light->targetent)
		{
			VectorSubtract (light->targetent->origin, light->origin, spotparms.spotvec);
			VectorNormalize (spotparms.spotvec);

			if (!light->angle)
				spotparms.falloff = -cos (20 * M_PI / 180);	
			else spotparms.falloff = -cos (light->angle / 2 * M_PI / 180);

			spotparms.isSpotLight = true;
		}
		else spotparms.isSpotLight = false;

		if (R_RenderLightAndShadow (&dl, (float) d_lightstylevalue[light->style], &spotparms))
			numlights++;
	}

	for (i = 0; i < MAX_DLIGHTS; i++)
	{
		if (cl_dlights[i].die < cl.time) continue;

		spotparms.isSpotLight = false;

		if (R_RenderLightAndShadow (&cl_dlights[i], 1.0f, &spotparms))
			numlights++;
	}

	glDisable (GL_SCISSOR_TEST);

	// if (numlights) Con_Printf ("%i lights\n", numlights);
}

