/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// sys_win.c -- Win32 system interface code

#include "quakedef.h"
#include "winquake.h"
#include "errno.h"
#include "resource.h"

#define MINIMUM_WIN_MEMORY		0x0880000
#define MAXIMUM_WIN_MEMORY		0x1000000

#define PAUSE_SLEEP		50				// sleep time on pause or minimization
#define NOT_FOCUS_SLEEP	20				// sleep time when not focus

qboolean	ActiveApp, Minimized;
qboolean	WinNT;

static qboolean		sc_return_on_enter = false;
HANDLE				hinput, houtput;

static char			*tracking_tag = "Clams & Mooses";

static HANDLE	hFile;
static HANDLE	heventParent;
static HANDLE	heventChild;

void MaskExceptions (void);
void Sys_PushFPCW_SetHigh (void);
void Sys_PopFPCW (void);


unsigned char zoneBuffer[64 * 1024 * 1024];

/*
===============================================================================

FILE IO

===============================================================================
*/

#define	MAX_HANDLES		10
FILE	*sys_handles[MAX_HANDLES];

int		findhandle (void)
{
	int		i;

	for (i = 1; i < MAX_HANDLES; i++)
		if (!sys_handles[i])
			return i;
	Sys_Error ("out of handles");
	return -1;
}

/*
================
filelength
================
*/
int filelength (FILE *f)
{
	int		pos;
	int		end;

	pos = ftell (f);
	fseek (f, 0, SEEK_END);
	end = ftell (f);
	fseek (f, pos, SEEK_SET);

	return end;
}

int Sys_FileOpenRead (char *path, int *hndl)
{
	FILE	*f;
	int		i, retval;

	i = findhandle ();

	f = fopen (path, "rb");

	if (!f)
	{
		*hndl = -1;
		retval = -1;
	}
	else
	{
		sys_handles[i] = f;
		*hndl = i;
		retval = filelength (f);
	}

	return retval;
}

int Sys_FileOpenWrite (char *path)
{
	FILE	*f;
	int		i;

	i = findhandle ();

	f = fopen (path, "wb");
	if (!f)
		Sys_Error ("Error opening %s: %s", path, strerror (errno));
	sys_handles[i] = f;

	return i;
}

void Sys_FileClose (int handle)
{
	fclose (sys_handles[handle]);
	sys_handles[handle] = NULL;
}

void Sys_FileSeek (int handle, int position)
{
	fseek (sys_handles[handle], position, SEEK_SET);
}

int Sys_FileRead (int handle, void *dest, int count)
{
	return fread (dest, 1, count, sys_handles[handle]);
}

int Sys_FileWrite (int handle, void *data, int count)
{
	return fwrite (data, 1, count, sys_handles[handle]);
}

int	Sys_FileTime (char *path)
{
	FILE	*f;
	int		retval;

	f = fopen (path, "rb");

	if (f)
	{
		fclose (f);
		retval = 1;
	}
	else
	{
		retval = -1;
	}

	return retval;
}

void Sys_mkdir (char *path)
{
	_mkdir (path);
}


/*
================
Sys_Init
================
*/
void Sys_Init (void)
{
	OSVERSIONINFO	vinfo;

	vinfo.dwOSVersionInfoSize = sizeof (vinfo);

	if (!GetVersionEx (&vinfo))
		Sys_Error ("Couldn't get OS info");

	if ((vinfo.dwMajorVersion < 4) || (vinfo.dwPlatformId == VER_PLATFORM_WIN32s))
		Sys_Error ("WinQuake requires at least Win95 or NT 4.0");

	if (vinfo.dwPlatformId == VER_PLATFORM_WIN32_NT)
		WinNT = true;
	else WinNT = false;
}


void Sys_Error (char *error, ...)
{
	va_list		argptr;
	char		text[1024], text2[1024];
	char		*text3 = "Press Enter to exit\n";
	char		*text4 = "***********************************\n";
	char		*text5 = "\n";
	DWORD		dummy;
	double		starttime;
	static int	in_sys_error0 = 0;
	static int	in_sys_error1 = 0;
	static int	in_sys_error2 = 0;
	static int	in_sys_error3 = 0;

	if (!in_sys_error3)
	{
		in_sys_error3 = 1;
	}

	va_start (argptr, error);
	vsprintf (text, error, argptr);
	va_end (argptr);

	// switch to windowed so the message box is visible, unless we already
	// tried that and failed
	if (!in_sys_error0)
	{
		in_sys_error0 = 1;
		VID_SetDefaultMode ();
		MessageBox (NULL, text, "Quake Error", MB_OK | MB_SETFOREGROUND | MB_ICONSTOP);
	}
	else
	{
		MessageBox (NULL, text, "Double Quake Error", MB_OK | MB_SETFOREGROUND | MB_ICONSTOP);
	}

	if (!in_sys_error1)
	{
		in_sys_error1 = 1;
		Host_Shutdown ();
	}

	// shut down QHOST hooks if necessary
	if (!in_sys_error2)
	{
		in_sys_error2 = 1;
	}

	exit (1);
}


void Sys_Quit (void)
{
	Host_Shutdown ();
	exit (0);
}


/*
================
Sys_DoubleTime
================
*/
double Sys_DoubleTime (void)
{
	static qboolean first = true;
	static __int64 qpcstart = 0;
	static __int64 qpcfreq = 0;
	__int64 qpcnow = 0;

	if (first)
	{
		first = false;

		// get higher resolution for the Sleep timer
		timeBeginPeriod (1);

		QueryPerformanceCounter ((LARGE_INTEGER *) &qpcstart);
		QueryPerformanceFrequency ((LARGE_INTEGER *) &qpcfreq);

		return 0;
	}

	QueryPerformanceCounter ((LARGE_INTEGER *) &qpcnow);

	return (double) (qpcnow - qpcstart) / (double) qpcfreq;
}


void Sys_Sleep (void)
{
	Sleep (1);
}


void Sys_SendKeyEvents (void)
{
	MSG        msg;

	while (PeekMessage (&msg, NULL, 0, 0, PM_NOREMOVE))
	{
		// we always update if there are any event, even if we're paused
		scr_skipupdate = 0;

		if (!GetMessage (&msg, NULL, 0, 0))
			Sys_Quit ();

		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
}


/*
==============================================================================

WINDOWS CRAP

==============================================================================
*/


/*
==================
WinMain
==================
*/
void SleepUntilInput (int time)
{
	MsgWaitForMultipleObjects (0, NULL, FALSE, time, QS_ALLINPUT);
}


/*
==================
WinMain
==================
*/
HINSTANCE	global_hInstance;
int			global_nCmdShow;
char		*argv[MAX_NUM_ARGVS];
static char	*empty_string = "";


// http://ntcoder.com/bab/tag/getuserprofiledirectory/
#include "userenv.h"
#pragma comment (lib, "userenv.lib")

char *Sys_GetUserHomeDir (void)
{
	static char szHomeDirBuf[MAX_PATH] = {0};

	// We need a process with query permission set
	HANDLE hToken = 0;
	DWORD BufSize = MAX_PATH;

	OpenProcessToken (GetCurrentProcess (), TOKEN_QUERY, &hToken);

	// Returns a path like C:/Documents and Settings/nibu if my user name is nibu
	GetUserProfileDirectory (hToken, szHomeDirBuf, &BufSize);

	// Close handle opened via OpenProcessToken
	CloseHandle (hToken);

	return szHomeDirBuf;
}


BOOL Sys_DirectoryExists (LPCTSTR szPath)
{
	DWORD dwAttrib = GetFileAttributes (szPath);
	return (dwAttrib != INVALID_FILE_ATTRIBUTES && (dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}


BOOL Sys_FileExists (LPCTSTR szPath)
{
	DWORD dwAttrib = GetFileAttributes (szPath);
	return (dwAttrib != INVALID_FILE_ATTRIBUTES && !(dwAttrib & FILE_ATTRIBUTE_DIRECTORY));
}


void Sys_SetWorkingDirectory (void)
{
	int i, j, k;

	char *homeDir = Sys_GetUserHomeDir ();

	const char *testDirs[] = {
		// these are the locations I have Quake installed to on various PCs; you may want to change it yourself
		homeDir,
		"\\Desktop Crap",
		"\\Games",
		"",
		NULL
	};

	const char *qDirs[] = {
		"Quake",
		"Q1",
		NULL
	};

	const char *qFiles[] = {
		"pak0.pak",
		"progs.dat",
		"config.cfg",
		NULL
	};

	// try to find ID1 content in the test paths
	for (i = 0;; i++)
	{
		if (!testDirs[i]) break;
		if (!Sys_DirectoryExists (testDirs[i])) continue;

		for (j = 0;; j++)
		{
			if (!qDirs[j]) break;
			if (!Sys_DirectoryExists (va ("%s\\%s\\"GAMENAME, testDirs[i], qDirs[j]))) continue;

			for (k = 0;; k++)
			{
				if (!qFiles[k]) break;
				if (!Sys_FileExists (va ("%s\\%s\\"GAMENAME"\\%s", testDirs[i], qDirs[j], qFiles[k]))) continue;

				SetCurrentDirectory (va ("%s\\%s", testDirs[i], qDirs[j]));
				return;
			}
		}
	}
}


int WINAPI WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	MSG				msg;
	quakeparms_t	parms;
	MEMORYSTATUS	lpBuffer;
	static	char	cwd[1024];
	int				t;
	RECT			rect;
	char			*realCmdLine;

#ifdef _DEBUG
	char debugCmdLine[256];

	strcpy (debugCmdLine, "-width 1024 -height 600 -window -heapsize 64000 -gamma 0.7");
	//strcpy (debugCmdLine, "-width 1366 -height 768 -heapsize 64000 -gamma 0.7");
	//strcpy (debugCmdLine, "-width 1440 -height 900 -heapsize 64000 -gamma 0.7");
	//strcpy (debugCmdLine, "-width 1920 -height 1080 -heapsize 64000 -gamma 0.7");
	realCmdLine = debugCmdLine;

	// force a working directory for debug builds because the exe isn't going to be in the correct game path
	// as an alternative we could mklink the quake gamedirs to the linker output path as a post-build step
	Sys_SetWorkingDirectory ();
#else
	realCmdLine = lpCmdLine;
#endif

	/* previous instances do not exist in Win32 */
	if (hPrevInstance)
		return 0;

	global_hInstance = hInstance;
	global_nCmdShow = nCmdShow;

	lpBuffer.dwLength = sizeof (MEMORYSTATUS);
	GlobalMemoryStatus (&lpBuffer);

	if (!GetCurrentDirectory (sizeof (cwd), cwd))
		Sys_Error ("Couldn't determine current directory");

	if (cwd[Q_strlen (cwd) - 1] == '/')
		cwd[Q_strlen (cwd) - 1] = 0;

	parms.basedir = cwd;
	parms.cachedir = NULL;

	parms.argc = 1;
	argv[0] = empty_string;

	while (*realCmdLine && (parms.argc < MAX_NUM_ARGVS))
	{
		while (*realCmdLine && ((*realCmdLine <= 32) || (*realCmdLine > 126)))
			realCmdLine++;

		if (*realCmdLine)
		{
			argv[parms.argc] = realCmdLine;
			parms.argc++;

			while (*realCmdLine && ((*realCmdLine > 32) && (*realCmdLine <= 126)))
				realCmdLine++;

			if (*realCmdLine)
			{
				*realCmdLine = 0;
				realCmdLine++;
			}
		}
	}

	parms.argv = argv;

	COM_InitArgv (parms.argc, parms.argv);

	parms.argc = com_argc;
	parms.argv = com_argv;

	parms.membase = zoneBuffer;
	parms.memsize = sizeof (zoneBuffer);

	Sys_Init ();

	// because sound is off until we become active
	S_BlockSound ();

	Host_Init (&parms);

	while (1)
	{
		// yield the CPU for a little while when paused, minimized, or not the focus
		if ((cl.paused && (!ActiveApp && !DDActive)) || Minimized || block_drawing)
		{
			SleepUntilInput (PAUSE_SLEEP);
			scr_skipupdate = 1;		// no point in bothering to draw
		}
		else if (!ActiveApp && !DDActive)
			SleepUntilInput (NOT_FOCUS_SLEEP);

		Host_Frame ();
	}

	/* return success of application */
	return TRUE;
}

